extern void *xm_memcpy(void *to, const void *from, size_t len);
extern void *ym_memcpy(void *to, const void *from, size_t len);
extern void *sse_memcpy(void *to, const void *from, size_t len);
extern void *memcpy_64(void *to, const void *from, size_t len);
extern void *memcpy1_64(void *to, const void *from, size_t len);

#define err(fmt, args...) \
	fprintf(stderr, "%s: Error: " fmt "\n", __func__, ##args)

#ifdef DEBUG
#define dbg(fmt, args...) \
	printf("%s: " fmt "\n", __func__, ##args)
#else
#define dbg(fmt, args...) \
	do { } while (0)
#endif

#define barrier() __asm__ __volatile__("": : :"memory")
#define mb()	  __asm__ __volatile__("mfence":::"memory")

typedef unsigned long long u64;
typedef unsigned char u8;

#ifdef __x86_64__
#define DECLARE_ARGS(val, low, high)    unsigned low, high
#define EAX_EDX_VAL(val, low, high)     ((low) | ((u64)(high) << 32))
#define EAX_EDX_ARGS(val, low, high)    "a" (low), "d" (high)
#define EAX_EDX_RET(val, low, high)     "=a" (low), "=d" (high)
#else
#define DECLARE_ARGS(val, low, high)    unsigned long long val
#define EAX_EDX_VAL(val, low, high)     (val)
#define EAX_EDX_ARGS(val, low, high)    "A" (val)
#define EAX_EDX_RET(val, low, high)     "=A" (val)
#endif

static double avg_xm, avg_mm;

static inline unsigned long long __rdtscll(void)
{
	DECLARE_ARGS(val, low, high);

	asm volatile("cpuid; rdtsc" : EAX_EDX_RET(val, low, high));

	return EAX_EDX_VAL(val, low, high);
}

#define rdtscll(val) do { (val) = __rdtscll(); } while (0)

#define rdtod(val)                                      \
do {                                                    \
        struct timeval tv;                              \
                                                        \
        gettimeofday(&tv, NULL);                        \
        (val) = tv.tv_sec * 1000000ULL + tv.tv_usec;    \
} while (0)
