#include <stdio.h>

#include "memcpy.h"

void *sse_memcpy(void *to, const void *from, size_t len)
{
	unsigned long src = (unsigned long)from;
	unsigned long dst = (unsigned long)to;
	void *p = to;
	int i;

	dbg("entry: src: %p, dst: %p, len: %lu", from, to, len);

	/* check alignment */
	if ((src ^ dst) & 0xf)
		goto unaligned;

	if (src & 0xf) {
		u8 chunk = 0x10 - (src & 0xf);

		dbg("Size of chunk to copy: %d, align: 0x%lx", chunk, src + chunk);

		/*
		 * copy chunk until 16-byte alignment
		 */
		memcpy_64(to, from, chunk);
		len -= chunk;
		to += chunk;
		from += chunk;
	}

	dbg("aligned copy: src: %p, dst: %p, len: %lu", from, to, len);

	/*
	 * copy in 256 Byte portions
	 */
	for (i = 0; i < (len & ~0xff); i+=256) {
		asm volatile(
		"movaps 0x0(%0),  %%xmm0\n\t"
		"movaps 0x10(%0), %%xmm1\n\t"
		"movaps 0x20(%0), %%xmm2\n\t"
		"movaps 0x30(%0), %%xmm3\n\t"
		"movaps 0x40(%0), %%xmm4\n\t"
		"movaps 0x50(%0), %%xmm5\n\t"
		"movaps 0x60(%0), %%xmm6\n\t"
		"movaps 0x70(%0), %%xmm7\n\t"
		"movaps 0x80(%0), %%xmm8\n\t"
		"movaps 0x90(%0), %%xmm9\n\t"
		"movaps 0xa0(%0), %%xmm10\n\t"
		"movaps 0xb0(%0), %%xmm11\n\t"
		"movaps 0xc0(%0), %%xmm12\n\t"
		"movaps 0xd0(%0), %%xmm13\n\t"
		"movaps 0xe0(%0), %%xmm14\n\t"
		"movaps 0xf0(%0), %%xmm15\n\t"

		"movaps %%xmm0,  0x0(%1)\n\t"
		"movaps %%xmm1,  0x10(%1)\n\t"
		"movaps %%xmm2,  0x20(%1)\n\t"
		"movaps %%xmm3,  0x30(%1)\n\t"
		"movaps %%xmm4,  0x40(%1)\n\t"
		"movaps %%xmm5,  0x50(%1)\n\t"
		"movaps %%xmm6,  0x60(%1)\n\t"
		"movaps %%xmm7,  0x70(%1)\n\t"
		"movaps %%xmm8,  0x80(%1)\n\t"
		"movaps %%xmm9,  0x90(%1)\n\t"
		"movaps %%xmm10, 0xa0(%1)\n\t"
		"movaps %%xmm11, 0xb0(%1)\n\t"
		"movaps %%xmm12, 0xc0(%1)\n\t"
		"movaps %%xmm13, 0xd0(%1)\n\t"
		"movaps %%xmm14, 0xe0(%1)\n\t"
		"movaps %%xmm15, 0xf0(%1)\n\t"
		:: "r" (from), "r" (to) : "memory");

		from += 256;
		to += 256;
	}

	goto trailer;

unaligned:
	/*
	 * copy in 256 Byte portions unaligned
	 */
	for (i = 0; i < (len & ~0xff); i+=256) {
		asm volatile(
		"movups 0x0(%0),  %%xmm0\n\t"
		"movups 0x10(%0), %%xmm1\n\t"
		"movups 0x20(%0), %%xmm2\n\t"
		"movups 0x30(%0), %%xmm3\n\t"
		"movups 0x40(%0), %%xmm4\n\t"
		"movups 0x50(%0), %%xmm5\n\t"
		"movups 0x60(%0), %%xmm6\n\t"
		"movups 0x70(%0), %%xmm7\n\t"
		"movups 0x80(%0), %%xmm8\n\t"
		"movups 0x90(%0), %%xmm9\n\t"
		"movups 0xa0(%0), %%xmm10\n\t"
		"movups 0xb0(%0), %%xmm11\n\t"
		"movups 0xc0(%0), %%xmm12\n\t"
		"movups 0xd0(%0), %%xmm13\n\t"
		"movups 0xe0(%0), %%xmm14\n\t"
		"movups 0xf0(%0), %%xmm15\n\t"

		"movups %%xmm0,  0x0(%1)\n\t"
		"movups %%xmm1,  0x10(%1)\n\t"
		"movups %%xmm2,  0x20(%1)\n\t"
		"movups %%xmm3,  0x30(%1)\n\t"
		"movups %%xmm4,  0x40(%1)\n\t"
		"movups %%xmm5,  0x50(%1)\n\t"
		"movups %%xmm6,  0x60(%1)\n\t"
		"movups %%xmm7,  0x70(%1)\n\t"
		"movups %%xmm8,  0x80(%1)\n\t"
		"movups %%xmm9,  0x90(%1)\n\t"
		"movups %%xmm10, 0xa0(%1)\n\t"
		"movups %%xmm11, 0xb0(%1)\n\t"
		"movups %%xmm12, 0xc0(%1)\n\t"
		"movups %%xmm13, 0xd0(%1)\n\t"
		"movups %%xmm14, 0xe0(%1)\n\t"
		"movups %%xmm15, 0xf0(%1)\n\t"
		:: "r" (from), "r" (to) : "memory");

		from += 256;
		to += 256;
	}

trailer:
	memcpy_64(to, from, len & 0xff);

	return p;
}
