#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <malloc.h>

#include "memcpy.h"
static void clflush(void *m)
{
	asm volatile("clflush (%0)" :: "r"(m));
}

static void align_on(void)
{
	long blah;
	//asm volatile("pushf\n\torq $0x40000, 0(%%rsp)\n\tpopf" : "=r"(blah));
}

static void align_off(void)
{
	long blah;
	asm volatile("pushf\n\tandq $~0x40000, 0(%%rsp)\n\tpopf" : "=r"(blah));
}

static void hexdump(const char *a, size_t sz, const char *msg)
{
	int i;

	printf("%s:\n", msg);

	for (i = 0; i < sz; i++) {

		if (i && !(i & 0xf))
			printf("\n");
		else if (i && !(i & 0x7))
			printf("  ");

		printf("%.2hhx ", a[i]);
	}
	printf("\n");
}

static inline int sanity_check_buffers(char *dst, char *src, unsigned size)
{
	int i = 0;

	if (!strncmp(src, dst, size))
		return 0;

	err("src <-> dst mismatch!\n");

	while (i < size) {
		char s = src[i];
		char d = dst[i];

		if (s != d)
			printf("Mismatch at %d: %.2hhx <-> %.2hhx\n", i, s, d);
		i++;
	}
	return 1;
}

static void bench_memcpy(char *dst, char *src, int srcmisalign, int dstmisalign, unsigned size, int loops)
{
	u64 xm_cycles, mm_cycles, tmp;
	int i;
	dst += dstmisalign;
	src += srcmisalign;

	avg_xm = avg_mm = 0;

	for (i = 0; i < loops; i++) {
		mb();
		align_on();
		rdtscll(tmp);
		clflush(src);
		clflush(dst);
		mb();

		ym_memcpy(dst, src, size);

		mb();
		rdtscll(xm_cycles);
		align_off();
		mb();

		if (sanity_check_buffers(dst, src, size))
			continue;

		xm_cycles -= tmp;
		avg_xm += xm_cycles;

		dbg("XM: size: %u, cycles: %llu\n", size, xm_cycles);

		mb();
		rdtscll(tmp);
		clflush(src);
		clflush(dst);
		mb();

		memcpy_64(dst, src, size);

		mb();
		rdtscll(mm_cycles);
		mb();

		mm_cycles -= tmp;
		avg_mm += mm_cycles;

		dbg("MM: size: %u, cycles: %llu\n", size, mm_cycles);
	}

	avg_xm /= loops;
	avg_mm /= loops;

	printf("%d(%d/%d)\t%g\t\t%g\t\t%.10g\n",
		size, srcmisalign, dstmisalign, avg_xm, avg_mm, avg_mm / avg_xm);
}

static void do_benchmark(unsigned size, unsigned loops, int rfd)
{
	char *src, *dst, *s1, *d1, *tmp;
	int i;

	size += 256;
	s1 = src = memalign(65536, size);
	if (!src)
		perror("malloc: src");

	d1 = dst = memalign(65536, size);
	if (!dst)
		perror("malloc: dst");
	//if (dst > src) { tmp = src; src = dst; dst = tmp; }

	if (read(rfd, src, size) != size)
		perror("urandom read");
	size -= 256;

	dbg("size: %d: src: %p, dst: %p",
	    size, src, dst);

	for (i = 0; i < 64; i += 4) {
		int j;
		for (j = i; j <= 64; j += 4) {
			bench_memcpy(dst, src, i, j, size, loops);
			if (j == i) continue;
			bench_memcpy(dst, src, j, i, size, loops);
		}
	}

#if 0
	hexdump(src, size, "Source buffer");
	hexdump(dst, size, "Dest buffer");
#endif

	free(s1);
	free(d1);
}

int main()
{
	int rfd, i, sz, loops = 100000;

	rfd = open("/dev/urandom", O_RDONLY);
	if (rfd < 0)
		perror("urandom open");

	printf("Benchmarking with %d iterations, average results:\n", loops);
	printf("size\tYM\t\tMM\t\tspeedup\n");

	for (i = 10; i < 16; i++) {
		sz = 1 << i;

		do_benchmark(sz, loops, rfd);
		do_benchmark(sz + sz/2, loops, rfd);
		do_benchmark(sz + sz/2 + sz/5, loops, rfd);
		do_benchmark(sz + sz/2 + sz/4, loops, rfd);
		do_benchmark(sz + sz/2 + sz/3, loops, rfd);
	}
	close(rfd);

	return 0;
}

