/*
 * xm_memcpy - SSSE3 version of memcpy
 *
 * Input:
 *  rdi destination
 *  rsi source
 *  rdx count
 *
 * Output:
 * rax original destination
 */
.globl xm_memcpy
.type xm_memcpy, @function

xm_memcpy:
	movq %rdi, %rax

	/*
	 * if ((src & 0xf) != (dst & 0xf))
	 *	goto unaligned;
	 */
	movzbq %sil, %r8
	movzbq %dil, %r9
	andb $0xf, %r8b
	andb $0xf, %r9b
	cmpb %r8b, %r9b
	jne .unaligned

	/* copy chunk of < 16 bytes up to beginning of aligned address */
	movzbq %sil, %rcx

	/* num bytes to copy */
	negb %cl

	/* prevent overflow when we're already aligned, i.e. %cl = 0 */
	andb $0xf, %cl
	subb %cl, %dl

	rep movsb

	/* num 16-byte blocks to copy */
	movq %rdx, %r8
	shrq $4, %r8

	.align 32
.repeat:
	cmpq $16, %r8
	jl .trailer

	movaps 0x0(%rsi), %xmm0
	movaps 0x10(%rsi), %xmm1
	movaps 0x20(%rsi), %xmm2
	movaps 0x30(%rsi), %xmm3
	movaps 0x40(%rsi), %xmm4
	movaps 0x50(%rsi), %xmm5
	movaps 0x60(%rsi), %xmm6
	movaps 0x70(%rsi), %xmm7
	movaps 0x80(%rsi), %xmm8
	movaps 0x90(%rsi), %xmm9
	movaps 0xa0(%rsi), %xmm10
	movaps 0xb0(%rsi), %xmm11
	movaps 0xc0(%rsi), %xmm12
	movaps 0xd0(%rsi), %xmm13
	movaps 0xe0(%rsi), %xmm14
	movaps 0xf0(%rsi), %xmm15

	movaps %xmm0, 0x0(%rdi)
	movaps %xmm1, 0x10(%rdi)
	movaps %xmm2, 0x20(%rdi)
	movaps %xmm3, 0x30(%rdi)
	movaps %xmm4, 0x40(%rdi)
	movaps %xmm5, 0x50(%rdi)
	movaps %xmm6, 0x60(%rdi)
	movaps %xmm7, 0x70(%rdi)
	movaps %xmm8, 0x80(%rdi)
	movaps %xmm9, 0x90(%rdi)
	movaps %xmm10, 0xa0(%rdi)
	movaps %xmm11, 0xb0(%rdi)
	movaps %xmm12, 0xc0(%rdi)
	movaps %xmm13, 0xd0(%rdi)
	movaps %xmm14, 0xe0(%rdi)
	movaps %xmm15, 0xf0(%rdi)

	/* advance pointers */
	addq $256, %rsi
	addq $256, %rdi
	subq $256, %rdx
	subq $16, %r8
	jnz .repeat
	jz .trailer

	.align 32
.unaligned:
	movq %rdx, %r8
	shrq $4, %r8

.repeat_ua:
	cmpq $16, %r8
	jl .trailer

	movups 0x0(%rsi), %xmm0
	movups 0x10(%rsi), %xmm1
	movups 0x20(%rsi), %xmm2
	movups 0x30(%rsi), %xmm3
	movups 0x40(%rsi), %xmm4
	movups 0x50(%rsi), %xmm5
	movups 0x60(%rsi), %xmm6
	movups 0x70(%rsi), %xmm7
	movups 0x80(%rsi), %xmm8
	movups 0x90(%rsi), %xmm9
	movups 0xa0(%rsi), %xmm10
	movups 0xb0(%rsi), %xmm11
	movups 0xc0(%rsi), %xmm12
	movups 0xd0(%rsi), %xmm13
	movups 0xe0(%rsi), %xmm14
	movups 0xf0(%rsi), %xmm15

	movups %xmm0, 0x0(%rdi)
	movups %xmm1, 0x10(%rdi)
	movups %xmm2, 0x20(%rdi)
	movups %xmm3, 0x30(%rdi)
	movups %xmm4, 0x40(%rdi)
	movups %xmm5, 0x50(%rdi)
	movups %xmm6, 0x60(%rdi)
	movups %xmm7, 0x70(%rdi)
	movups %xmm8, 0x80(%rdi)
	movups %xmm9, 0x90(%rdi)
	movups %xmm10, 0xa0(%rdi)
	movups %xmm11, 0xb0(%rdi)
	movups %xmm12, 0xc0(%rdi)
	movups %xmm13, 0xd0(%rdi)
	movups %xmm14, 0xe0(%rdi)
	movups %xmm15, 0xf0(%rdi)

	/* advance pointers */
	addq $256, %rsi
	addq $256, %rdi
	subq $256, %rdx
	subq $16, %r8
	jnz .repeat_ua

	.align 32
.trailer:
	movq %rdx, %rcx
	movq %rdx, %r8
	andq $0x7, %r8
	shrq $3, %rcx
	rep; movsq

	movq %r8, %rcx
	rep; movsb

	retq
